@@ -6,9 +6,8 @@
 #endif
 
 #include "lib/nls.h"
+#include "lib/utils_crypt.h"
 
-#define MAX_CIPHER_LEN		32
-#define MAX_CIPHER_LEN_STR	"32"
 #define DEFAULT_CIPHER(type)	(DEFAULT_##type##_CIPHER "-" DEFAULT_##type##_MODE)
 
 #define log_dbg(x...) clogger(NULL, CRYPT_LOG_DEBUG, __FILE__, __LINE__, x)
